/*
 *
 * @Component			OMAPCONF
 * @Filename			dep44xx.c
 * @Description			OMAP4 Clock Domain Static Dependencies
 *				Definitions & Functions
 * @Author			Patrick Titiano (p-titiano@ti.com)
 * @Date			2011
 * @Copyright			Texas Instruments Incorporated
 *
 *
 * Copyright (C) 2011 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <dep44xx.h>
#include <lib.h>
#include <mem.h>
#include <lib_android.h>
#include <prm44xx.h>
#include <cm44xx.h>
#include <clkdm44xx.h>
#include <cpuinfo.h>
#include <autoadjust_table.h>
#include <help.h>


/* #define DEP44XX_DEBUG */
#ifdef DEP44XX_DEBUG
#define dprintf(format, ...)	 printf(format, ## __VA_ARGS__)
#else
#define dprintf(format, ...)
#endif


typedef enum {
	DEP_STATIC,
	DEP_DYNAMIC,
	DEP44XX_TYPE_MAX
} dep44xx_type;



#define DEP_NA		0 /* not available, does not exist */
#define DEP_RW		1 /* available, SW configurable */
#define DEP_RO		2 /* available, not configurable (hard-coded) */

#define DEP_DIS		3 /* disabled */
#define DEP_EN		4 /* enabled */


/*
 * Table used to retrieve clock domain dependency register.
 * 0 = does not exist
 */
static const unsigned int dep44xx_reg_table[OMAP4_CD_ID_MAX][2] = {
	{0, 0}, /* WKUP */
	{0, OMAP4430_CM_EMU_DYNAMICDEP}, /* EMU */
	{OMAP4430_CM_MPU_STATICDEP, OMAP4430_CM_MPU_DYNAMICDEP}, /* MPU */
	{0, 0}, /* ABE */
	{OMAP4430_CM_IVAHD_STATICDEP, OMAP4430_CM_IVAHD_DYNAMICDEP}, /* IVAHD */
	{OMAP4430_CM_DSP_STATICDEP, OMAP4430_CM_DSP_DYNAMICDEP}, /* DSP */
	{0, 0}, /* AO_L4 */
	{0, OMAP4430_CM_L4CFG_DYNAMICDEP}, /* L4CFG */
	{OMAP4430_CM_C2C_STATICDEP, OMAP4430_CM_C2C_DYNAMICDEP}, /* C2C */
	{OMAP4430_CM_SDMA_STATICDEP, OMAP4430_CM_SDMA_DYNAMICDEP}, /* SDMA */
	{OMAP4430_CM_MPU_M3_STATICDEP, OMAP4430_CM_MPU_M3_DYNAMICDEP}, /* MPU_M3 */
	{0, OMAP4430_CM_L3_1_DYNAMICDEP}, /* L3_1 */
	{0, OMAP4430_CM_L3_2_DYNAMICDEP}, /* L3_2 */
	{0, 0}, /* L3_INSTR */
	{0, 0}, /* EMIF */
	{0, 0}, /* STD_EFUSE */
	{0, 0}, /* CUST_EFUSE */
	{OMAP4430_CM_CAM_STATICDEP, OMAP4430_CM_CAM_DYNAMICDEP}, /* CAM */
	{OMAP4430_CM_DSS_STATICDEP, OMAP4430_CM_DSS_DYNAMICDEP}, /* DSS */
	{OMAP4430_CM_GFX_STATICDEP, OMAP4430_CM_GFX_DYNAMICDEP}, /* GFX */
	{OMAP4430_CM_L3INIT_STATICDEP, OMAP4430_CM_L3INIT_DYNAMICDEP}, /* L3INIT */
	{OMAP4430_CM_L4SEC_STATICDEP, OMAP4430_CM_L4SEC_DYNAMICDEP}, /* L4SEC */
	{0, OMAP4430_CM_L4PER_DYNAMICDEP}, /* L4PER */
	{0, 0} };  /* NONE */


/*
 * Table used to retrieve dependency bit position in dependency register
 * -1 = does not exist
 */
static const short int dep44xx_bit_pos_table[OMAP4_CD_ID_MAX] = {
	15, /* WKUP */
	-1, /* EMU */
	-1, /* MPU */
	3, /* ABE */
	2, /* IVAHD */
	1, /* DSP */
	16, /* AO_L4 */
	12, /* L4CFG */
	18, /* C2C */
	11, /* SDMA */
	0, /* MPU_M3 */
	5, /* L3_1 */
	6, /* L3_2 */
	-1, /* L3_INSTR */
	4, /* EMIF */
	-1, /* STD_EFUSE */
	17, /* CUST_EFUSE */
	9, /* CAM */
	8, /* DSS */
	10, /* GFX */
	7, /* L3INIT */
	14, /* L4SEC */
	13, /* L4PER */
	-1};  /* NONE */


/*
 * Table to be used to determine if a static dependency between 2 clock domains
 * exists or not (i.e there is an associated control bit in CM_xyz_STATICDEP
 * register.
 * 0 = does not exist
 * 1 = present
 * 2 = present, read-only (hard-wired)
 */
static const unsigned int statdep4430_table[OMAP4_CD_ID_MAX][OMAP4_CD_ID_MAX] = {
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* WKUP */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* EMU */
	{1, 0, 0, 1, 1, 1, 2, 1, 2, 2, 1, 1, 1, 0, 1, 0, 2, 2, 1, 1, 1, 1, 1, 0}, /* MPU */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* ABE */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* IVAHD */
	{1, 0, 0, 1, 1, 0, 2, 1, 0, 0, 0, 1, 1, 0, 1, 0, 2, 2, 0, 0, 1, 0, 1, 0}, /* DSP */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* AO_L4 */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L4CFG */
	{0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0}, /* C2C */
	{1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 2, 0, 1, 0, 0, 2, 1, 0, 1, 1, 1, 0}, /* SDMA */
	{1, 0, 0, 1, 1, 1, 2, 1, 0, 2, 0, 1, 1, 0, 1, 0, 2, 2, 1, 1, 1, 1, 1, 0}, /* MPU_M3 */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L3_1 */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L3_2 */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L3_INSTR */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* EMIF */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* STD_EFUSE */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* CUST_EFUSE */
	{0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* CAM */
	{0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 2, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* DSS */
	{0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* GFX */
	{1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 2, 2, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0}, /* L3INIT */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0}, /* L4SEC */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L4PER */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} };  /* NONE */


/*
 * Table to be used to determine if a static dependency between 2 clock domains
 * exists or not (i.e there is an associated control bit in CM_xyz_STATICDEP
 * register.
 * 0 = does not exist
 * 1 = present
 * 2 = present, read-only (hard-wired)
 */
static const unsigned int statdep4470_table[OMAP4_CD_ID_MAX][OMAP4_CD_ID_MAX] = {
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* WKUP */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* EMU */
	{1, 0, 0, 1, 1, 1, 2, 1, 2, 2, 1, 1, 1, 0, 1, 0, 2, 2, 1, 1, 1, 1, 1, 0}, /* MPU */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* ABE */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* IVAHD */
	{1, 0, 0, 1, 1, 0, 2, 1, 0, 0, 0, 1, 1, 0, 1, 0, 2, 2, 0, 0, 1, 0, 1, 0}, /* DSP */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* AO_L4 */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L4CFG */
	{0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0}, /* C2C */
	{1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 2, 0, 1, 0, 0, 2, 1, 0, 1, 1, 1, 0}, /* SDMA */
	{1, 0, 0, 1, 1, 1, 2, 1, 0, 2, 0, 1, 1, 0, 1, 0, 2, 2, 1, 1, 1, 1, 1, 0}, /* MPU_M3 */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L3_1 */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L3_2 */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L3_INSTR */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* EMIF */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* STD_EFUSE */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* CUST_EFUSE */
	{0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* CAM */
	{0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 2, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* DSS */
	{0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* GFX */
	{1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 2, 2, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0}, /* L3INIT */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0}, /* L4SEC */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, /* L4PER */
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} };  /* NONE */


/*
 * Table to be used to store expected (Plan of Record)
 * static dependency configuration, between 2 clock domains.
 */
static const unsigned int por_statdep4430_gb_table[OMAP4_CD_ID_MAX][OMAP4_CD_ID_MAX] = {
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* WKUP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMU */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_EN, DEP_NA, DEP_EN, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_DIS, DEP_EN, DEP_NA}, /* MPU */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* ABE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* IVAHD */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* DSP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* AO_L4 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4CFG */
	{DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* CDEP_DISC */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA}, /* SDMA */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA}, /* MPU_M3 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_INSTR */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMIF */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* STD_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CUST_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CAM */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* DSS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* GFX */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA}, /* L3INIT */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA}, /* L4SEC */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4PER */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA} };  /* NONE */

static const unsigned int por_statdep4430_ics_table[OMAP4_CD_ID_MAX][OMAP4_CD_ID_MAX] = {
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* WKUP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMU */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_EN, DEP_NA, DEP_EN, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_NA}, /* MPU */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* ABE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* IVAHD */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* DSP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* AO_L4 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4CFG */
	{DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* CDEP_DISC */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA}, /* SDMA */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_NA}, /* MPU_M3 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_INSTR */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMIF */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* STD_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CUST_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CAM */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* DSS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* GFX */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA}, /* L3INIT */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA}, /* L4SEC */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4PER */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA} };  /* NONE */

static const unsigned int por_statdep4460_gb_table[OMAP4_CD_ID_MAX][OMAP4_CD_ID_MAX] = {
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* WKUP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMU */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_DIS, DEP_EN, DEP_NA}, /* MPU */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* ABE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* IVAHD */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* DSP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* AO_L4 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4CFG */
	{DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* CDEP_DISC */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA}, /* SDMA */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA}, /* MPU_M3 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_INSTR */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMIF */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* STD_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CUST_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CAM */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* DSS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* GFX */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA}, /* L3INIT */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA}, /* L4SEC */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4PER */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA} };  /* NONE */

static const unsigned int por_statdep4460_ics_table[OMAP4_CD_ID_MAX][OMAP4_CD_ID_MAX] = {
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* WKUP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMU */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_NA}, /* MPU */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* ABE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* IVAHD */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* DSP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* AO_L4 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4CFG */
	{DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* CDEP_DISC */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA}, /* SDMA */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_NA}, /* MPU_M3 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_INSTR */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMIF */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* STD_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CUST_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CAM */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* DSS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* GFX */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA}, /* L3INIT */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA}, /* L4SEC */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4PER */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA} };  /* NONE */

static const unsigned int por_statdep4470_gb_table[OMAP4_CD_ID_MAX][OMAP4_CD_ID_MAX] = {
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* WKUP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMU */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_DIS, DEP_EN, DEP_NA}, /* MPU */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* ABE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* IVAHD */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* DSP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* AO_L4 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4CFG */
	{DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* CDEP_DISC */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA}, /* SDMA */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA}, /* MPU_M3 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_INSTR */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMIF */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* STD_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CUST_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CAM */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* DSS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* GFX */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA}, /* L3INIT */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA}, /* L4SEC */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4PER */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA} };  /* NONE */

static const unsigned int por_statdep4470_ics_table[OMAP4_CD_ID_MAX][OMAP4_CD_ID_MAX] = {
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* WKUP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMU */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_NA}, /* MPU */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* ABE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* IVAHD */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* DSP */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* AO_L4 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4CFG */
	{DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA}, /* CDEP_DISC */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_NA}, /* SDMA */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_EN, DEP_EN, DEP_NA, DEP_DIS, DEP_NA, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_DIS, DEP_EN, DEP_NA}, /* MPU_M3 */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_DEP_DIS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L3_INSTR */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* EMIF */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* STD_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CUST_EFUSE */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* CAM */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* DSS */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* GFX */
	{DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA}, /* L3INIT */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_DIS, DEP_NA, DEP_DIS, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_DIS, DEP_NA}, /* L4SEC */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA}, /* L4PER */
	{DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA, DEP_NA} };  /* NONE */


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_get
 * @BRIEF		return static dependency from 'from' clock domain
 *			towards 'to' domain.
 * @RETURNS		0 = does not exist
 *			1 = present
 *			2 = present, read-only (hard-wired)
 *			OMAPCONF_ERR_ARG
 * @param[in]		from: master clock domain ID
 * @param[in]		to: target clock domain ID
 * @DESCRIPTION		return static dependency from 'from' clock domain
 *			towards 'to' domain.
 *//*------------------------------------------------------------------------ */
int statdep44xx_get(clkdm44xx_id from, clkdm44xx_id to)
{
	CHECK_ARG_LESS_THAN(from, OMAP4_CD_ID_MAX, OMAPCONF_ERR_ARG);
	CHECK_ARG_LESS_THAN(to, OMAP4_CD_ID_MAX, OMAPCONF_ERR_ARG);

	if (cpu_is_omap4470())
		return (int) statdep4470_table[from][to];
	else
		return (int) statdep4430_table[from][to];
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_por_cfg_get
 * @BRIEF		return expected static dependency configuration from
 *			'from' clock domain towards 'to' domain.
 * @RETURNS		DEP_DIS = disabled
 *			DEP_EN = enabled
 *			DEP_NA = does not exist
 *			OMAPCONF_ERR_ARG
 *			OMAPCONF_ERR_CPU
 * @param[in]		from: master clock domain ID
 * @param[in]		to: target clock domain ID
 * @DESCRIPTION		return expected static dependency configuration from
 *			'from' clock domain towards 'to' domain.
 *//*------------------------------------------------------------------------ */
int statdep44xx_por_cfg_get(clkdm44xx_id from, clkdm44xx_id to)
{
	android_pastry_id pastry;

	CHECK_CPU(44xx, OMAPCONF_ERR_CPU);
	CHECK_ARG_LESS_THAN(from, OMAP4_CD_ID_MAX, OMAPCONF_ERR_ARG);
	CHECK_ARG_LESS_THAN(to, OMAP4_CD_ID_MAX, OMAPCONF_ERR_ARG);

	pastry = android_pastry_get();

	/* FIXME: find expected OPP for Generic Linux */
	if (cpu_is_omap4430()) {
		if (pastry < PASTRY_ICS)
			return (int) por_statdep4430_gb_table[from][to];
		else
			return (int) por_statdep4430_ics_table[from][to];
	} else if (cpu_is_omap4460()) {
		if (pastry < PASTRY_ICS)
			return (int) por_statdep4460_gb_table[from][to];
		else
			return (int) por_statdep4460_ics_table[from][to];
	} else {
		if (pastry < PASTRY_ICS)
			return (int) por_statdep4470_gb_table[from][to];
		else
			return (int) por_statdep4470_ics_table[from][to];
	}
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		dep44xx_register_get
 * @BRIEF		return the dependency register physical address.
 * @RETURNS		0 in case of error
 *			register physical address otherwise
 * @param[in]		id: clock domain ID
 * @param[in]		type: static or dynamic
 * @DESCRIPTION		return the dependency register physical address.
 *//*------------------------------------------------------------------------ */
unsigned int dep44xx_register_get(clkdm44xx_id id, dep44xx_type type)
{
	unsigned int cm_dep_addr;
	#ifdef DEP44XX_DEBUG
	char id_s[CLKDM44XX_MAX_NAME_LENGTH];
	#endif

	if (id >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): id (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, id);
		return 0;
	}

	if (type >= DEP44XX_TYPE_MAX) {
		fprintf(stderr, "%s(): type (%d) >= DEP44XX_TYPE_MAX!!!\n",
			__func__, type);
		return 0;
	}

	cm_dep_addr = dep44xx_reg_table[id][type];
	#ifdef DEP44XX_DEBUG
	if (cm_dep_addr == 0) {
		/* Dependency does not exist */
		dprintf("%s(%s, %d): does not exist.\n",
		__func__, clkdm44xx_get_name(id, id_s), type);
	} else {
		dprintf("%s(%s, %d): dependency register address = 0x%08X\n",
		__func__, clkdm44xx_get_name(id, id_s), type, cm_dep_addr);
	}
	#endif

	return cm_dep_addr;
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		dep44xx_is_enabled
 * @BRIEF		check if a static dependency between 2 clock domains
 *			is enabled or not.
 * @RETURNS		1 if dependency is enabled
 *			0 if dependency is disabled
 *			OMAPCONF_ERR_NOT_AVAILABLE if dependency does not exist
 *			OMAPCONF_ERR_ARG
 *			OMAPCONF_ERR_REG_ACCESS
 * @param[in]		from: clock domain ID
 * @param[in]		to: clock domain ID
 * @param[in]		type: static or dynamic
 * @DESCRIPTION		check if a static dependency between 2 clock domains
 *			is enabled or not.
 *//*------------------------------------------------------------------------ */
int dep44xx_is_enabled(
	clkdm44xx_id from, clkdm44xx_id to, dep44xx_type type)
{
	unsigned int cm_dep_addr, cm_dep;
	int ret;
	#ifdef DEP44XX_DEBUG
	char to_s[CLKDM44XX_MAX_NAME_LENGTH];
	char from_s[CLKDM44XX_MAX_NAME_LENGTH];
	#endif

	if (!cpu_is_omap44xx())
		return OMAPCONF_ERR_CPU;

	if (from >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): from id (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, from);
		return OMAPCONF_ERR_ARG;
	}

	if (to >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): to id (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, to);
		return OMAPCONF_ERR_ARG;
	}

	if (type >= DEP44XX_TYPE_MAX) {
		fprintf(stderr, "%s(): type (%d) >= DEP44XX_TYPE_MAX!!!\n",
			__func__, type);
		return OMAPCONF_ERR_ARG;
	}

	/* Get dependency register */
	cm_dep_addr = dep44xx_register_get(from, type);
	if (cm_dep_addr == 0) {
		/* Dependency does not exist */
		dprintf("%s(%s, %s, %d): does not exist.\n", __func__,
			clkdm44xx_get_name(from, from_s),
			clkdm44xx_get_name(to, to_s),
			type);
		return OMAPCONF_ERR_NOT_AVAILABLE;
	}

	/* Read Register */
	ret = mem_read(cm_dep_addr, &cm_dep);
	if (ret != 0)
		return OMAPCONF_ERR_REG_ACCESS;
	dprintf("%s(%s, %s, %d): dependency register value = 0x%08X\n",
		__func__,
		clkdm44xx_get_name(from, from_s),
		clkdm44xx_get_name(to, to_s),
		type, cm_dep);

	/* Get Dependency setting from register */
	if (dep44xx_bit_pos_table[to] != -1) {
		ret = extract_bit(cm_dep, dep44xx_bit_pos_table[to]);
		dprintf("%s(%s, %s, %d): pos = %d, dependency = %d\n",
			__func__,
			clkdm44xx_get_name(from, from_s),
			clkdm44xx_get_name(to, to_s),
			type, dep44xx_bit_pos_table[to],
			extract_bit(cm_dep, dep44xx_bit_pos_table[to]));
	} else {
		dprintf("%s(%s, %s, %d): dependency does not exist.\n",
			__func__,
			clkdm44xx_get_name(from, from_s),
			clkdm44xx_get_name(to, to_s),
			type);
		ret = OMAPCONF_ERR_NOT_AVAILABLE;
	}

	return ret;
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_has_dependency_towards_it
 * @BRIEF		indicate if clock domain can be dependent on another one
 * @RETURNS		1 if clock domain is dependent on another one
 *			0 if clock domain is NOT dependent on another one
 *			OMAPCONF_ERR_ARG
 * @param[in]		id: clock domain ID
 * @DESCRIPTION		indicate if clock domain can be dependent on another one
 *//*------------------------------------------------------------------------ */
int statdep44xx_has_dependency_towards_it(clkdm44xx_id id)
{
	if (id >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): id (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, id);
		return OMAPCONF_ERR_ARG;
	}

	if (dep44xx_bit_pos_table[id] != -1)
		return 1;
	else
		return 0;
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		dep44xx_register_exists
 * @BRIEF		indicate if given domain has a dependency register
 * @RETURNS		1 given domain has a dependency register.
 *			0 given domain has NO dependency register.
 *			OMAPCONF_ERR_ARG
 * @param[in]		id: clock domain ID
 * @param[in]		type: static or dynamic
 * @DESCRIPTION		indicate if given domain has a dependency register
 *//*------------------------------------------------------------------------ */
int dep44xx_register_exists(clkdm44xx_id id, dep44xx_type type)
{
	#ifdef DEP44XX_DEBUG
	char id_s[CLKDM44XX_MAX_NAME_LENGTH];
	#endif

	if (id >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): id (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, id);
		return OMAPCONF_ERR_ARG;
	}
	if (type >= DEP44XX_TYPE_MAX) {
		fprintf(stderr, "%s(): type (%d) >= DEP44XX_TYPE_MAX!!!\n",
			__func__, type);
		return OMAPCONF_ERR_ARG;
	}

	if (dep44xx_reg_table[id][type] != 0) {
		dprintf("%s(%s) = 1 (has statdep)\n", __func__,
			clkdm44xx_get_name(id, id_s));
		return 1;
	} else {
		dprintf("%s(%s) = 0 (NO statdep)\n", __func__,
			clkdm44xx_get_name(id, id_s));
		return 0;
	}
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_exists
 * @BRIEF		indicate if there is a static dependency between the 2
 *			given clock domains.
 * @RETURNS		1 there is a static dependency between the
 *			2 clock domains
 *			0 there is NO static dependency between the
 *			2 clock domains
 *			OMAPCONF_ERR_ARG
 * @param[in]		from: clock domain ID
 * @param[in]		to: clock domain ID
 * @DESCRIPTION		indicate if there is a static dependency between the 2
 *			given clock domains.
 *//*------------------------------------------------------------------------ */
int statdep44xx_exists(clkdm44xx_id from, clkdm44xx_id to)
{
	#ifdef DEP44XX_DEBUG
	char to_s[CLKDM44XX_MAX_NAME_LENGTH];
	char from_s[CLKDM44XX_MAX_NAME_LENGTH];
	#endif

	if (from >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): from (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, from);
		return OMAPCONF_ERR_ARG;
	}
	if (to >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): to (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, to);
		return OMAPCONF_ERR_ARG;
	}

	if (statdep44xx_get(from, to) == 0) {
		dprintf("%s(%s, %s)=0 (statdep44xx_table[%d][%d]=%d)\n",
			__func__, clkdm44xx_get_name(from, from_s),
			clkdm44xx_get_name(to, to_s),
			from, to, statdep44xx_get(from, to));
		return 0;
	} else {
		dprintf("%s(%s, %s)=1 (statdep44xx_table[%d][%d]=%d)\n",
			__func__, clkdm44xx_get_name(from, from_s),
			clkdm44xx_get_name(to, to_s),
			from, to, statdep44xx_get(from, to));
		return 1;
	}
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_set
 * @BRIEF		clear/set given static dependency.
 * @RETURNS		OMAPCONF_ERR_CPU
 *			OMAPCONF_ERR_NOT_AVAILABLE if dependency does not exist
 *			OMAPCONF_ERR_ARG
 *			OMAPCONF_ERR_REG_ACCESS
 * @param[in]		from: clock domain ID
 * @param[in]		to: clock domain ID
 * @param[in]		enable: 0 to clear, 1 to set static dependency.
 * @DESCRIPTION		clear/set given static dependency.
 *//*------------------------------------------------------------------------ */
int statdep44xx_set(clkdm44xx_id from, clkdm44xx_id to, int enable)
{
	int ret;
	unsigned int cm_dep_addr, cm_dep;
	#ifdef DEP44XX_DEBUG
	char to_s[CLKDM44XX_MAX_NAME_LENGTH];
	char from_s[CLKDM44XX_MAX_NAME_LENGTH];
	#endif

	if (!cpu_is_omap44xx())
		return OMAPCONF_ERR_CPU;

	if (from >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): from id (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, from);
		return OMAPCONF_ERR_ARG;
	}

	if (to >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): to id (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, to);
		return OMAPCONF_ERR_ARG;
	}

	if (dep44xx_register_exists(from, DEP_STATIC) != 1) {
		ret = OMAPCONF_ERR_NOT_AVAILABLE;
		goto statdep44xx_set_end;
	}

	if (statdep44xx_has_dependency_towards_it(to) != 1) {
		ret = OMAPCONF_ERR_NOT_AVAILABLE;
		goto statdep44xx_set_end;
	}

	/* Get dependency register */
	cm_dep_addr = dep44xx_register_get(from, 0);
	if (cm_dep_addr == 0) {
		/* Dependency does not exist */
		dprintf("%s(%s, %s, %d): does not exist.\n", __func__,
			clkdm44xx_get_name(from, from_s),
			clkdm44xx_get_name(to, to_s),
			enable);
		ret = OMAPCONF_ERR_NOT_AVAILABLE;
		goto statdep44xx_set_end;
	}

	cm_dep_addr = dep44xx_reg_table[from][0];
	dprintf("%s(%s, %s, %d): dependency register address = 0x%08X\n",
		__func__,
		clkdm44xx_get_name(from, from_s),
		clkdm44xx_get_name(to, to_s),
		enable, cm_dep_addr);

	/* Read register content */
	ret = mem_read(cm_dep_addr, &cm_dep);
	if (ret != 0) {
		ret = OMAPCONF_ERR_REG_ACCESS;
		goto statdep44xx_set_end;
	}
	dprintf("%s(%s, %s, %d): dependency register value = 0x%08X\n",
		__func__,
		clkdm44xx_get_name(from, from_s),
		clkdm44xx_get_name(to, to_s),
		enable, cm_dep);

	/* Update register content */
	if (enable)
		cm_dep |= (1 << dep44xx_bit_pos_table[to]);
	else
		cm_dep &= ~(1 << dep44xx_bit_pos_table[to]);

	/* Save updated register content */
	ret = mem_write(cm_dep_addr, cm_dep);
	if (ret != 0) {
		ret = OMAPCONF_ERR_REG_ACCESS;
		goto statdep44xx_set_end;
	}

	ret = 0;

statdep44xx_set_end:
	return ret;
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_is_read_only
 * @BRIEF		indicate if an existing static dependency between 2
 *			given clock domains is SW configurable or hard-coded.
 * @RETURNS		1 static dependency is hard-coded
 *			0 static dependency is SW configurable
 *			OMAPCONF_ERR_NOT_AVAILABLE if static dependency does not
 *			exist
 *			OMAPCONF_ERR_ARG
 *			OMAPCONF_ERR_INTERNAL
 * @param[in]		from: clock domain ID
 * @param[in]		to: clock domain ID
 * @DESCRIPTION		indicate if an existing static dependency between 2
 *			given clock domains is SW configurable or hard-coded.
 *//*------------------------------------------------------------------------ */
int statdep44xx_is_read_only(clkdm44xx_id from, clkdm44xx_id to)
{
	if (from >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): from (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, from);
		return OMAPCONF_ERR_ARG;
	}
	if (to >= OMAP4_CD_ID_MAX) {
		fprintf(stderr, "%s(): to (%d) >= OMAP4_CD_ID_MAX!!!\n",
			__func__, to);
		return OMAPCONF_ERR_ARG;
	}

	switch (statdep44xx_get(from, to)) {
	case 0:
		return OMAPCONF_ERR_NOT_AVAILABLE;
	case 1:
		return 0;
	case 2:
		return 1;
	default:
		fprintf(stderr, "%s(): unexpected value in "
			"statdep44xx_table[%d][%d] (%d)!!!\n",
			__func__, from, to, statdep44xx_get(from, to));
		return OMAPCONF_ERR_INTERNAL;
	}

}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_dump
 * @BRIEF		dump all static dependency registers content
 * @RETURNS		0 in case of success
 *			OMAPCONF_ERR_CPU
 *			OMAPCONF_ERR_REG_ACCESS
 * @param[in]		none
 * @DESCRIPTION		dump all static dependency registers content
 *//*------------------------------------------------------------------------ */
int statdep44xx_dump(void)
{
	clkdm44xx_id id;
	char id_s[CLKDM44XX_MAX_NAME_LENGTH];
	char dep_reg_name[CLKDM44XX_MAX_NAME_LENGTH + 3 + 11 + 1];
	int ret;
	char table[TABLE_MAX_ROW][TABLE_MAX_COL][TABLE_MAX_ELT_LEN];
	unsigned int row, cm_dep_addr, cm_dep;

	/* Check CPU */
	if (!cpu_is_omap44xx())
		return OMAPCONF_ERR_CPU;

	row = 0;
	autoadjust_table_init(table);

	strncpy(table[row][0], "Reg. Name", TABLE_MAX_ELT_LEN);
	strncpy(table[row][1], "Reg. Addr",
		TABLE_MAX_ELT_LEN);
	strncpy(table[row][2], "Reg. Val.", TABLE_MAX_ELT_LEN);
	row++;

	/* Browse all clock domains */
	for (id = OMAP4_CD_WKUP; id < OMAP4_CD_NONE; id++) {
		if (dep44xx_register_exists(id, DEP_STATIC) != 1)
			continue;

		/* Store register name in table */
		strcpy(dep_reg_name , "CM_");
		clkdm44xx_get_name(id, id_s);
		strcat(dep_reg_name, id_s);
		strcat(dep_reg_name, "_STATICDEP");
		snprintf(table[row][0], TABLE_MAX_ELT_LEN, "%s", dep_reg_name);

		/* Get register address */
		cm_dep_addr = dep44xx_register_get(id, DEP_STATIC);
		/* Store register address in table */
		snprintf(table[row][1], TABLE_MAX_ELT_LEN, "0x%08X",
			cm_dep_addr);

		/* Read Register */
		ret = mem_read(cm_dep_addr, &cm_dep);
		if (ret != 0)
			return OMAPCONF_ERR_REG_ACCESS;
		/* Store register content in table */
		snprintf(table[row++][2], TABLE_MAX_ELT_LEN, "0x%08X", cm_dep);
	}

	autoadjust_table_print(table, row, 3);

	return 0;
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_show
 * @BRIEF		show complete static dependencies configuration
 * @RETURNS		0 in case of success
 *			OMAPCONF_ERR_CPU
 *			OMAPCONF_ERR_ARG
 *			OMAPCONF_ERR_REG_ACCESS
 * @param[in]
 * @DESCRIPTION		show complete static dependencies configuration
 *//*------------------------------------------------------------------------ */
int statdep44xx_show(void)
{
	clkdm44xx_id from, to;
	int ret;
	char table[TABLE_MAX_ROW][TABLE_MAX_COL][TABLE_MAX_ELT_LEN];
	unsigned int row, col, max_col;
	char to_s[CLKDM44XX_MAX_NAME_LENGTH];
	char from_s[CLKDM44XX_MAX_NAME_LENGTH];

	/* Check CPU */
	if (!cpu_is_omap44xx())
		return OMAPCONF_ERR_CPU;

	row = 0;
	autoadjust_table_init(table);

	/* Fill Table Header */
	max_col = 0;
	strncpy(table[row][0], "Static Dependency FROM", TABLE_MAX_ELT_LEN);
	for (from = OMAP4_CD_WKUP; from < OMAP4_CD_NONE; from++) {
		if (dep44xx_register_exists(from, DEP_STATIC) != 1)
			continue;

		max_col++;
		clkdm44xx_get_name(from, from_s);
		snprintf(table[row][max_col], TABLE_MAX_ELT_LEN, "%s", from_s);
	}
	row++;
	strncpy(table[row][0], "TOWARDS", TABLE_MAX_ELT_LEN);
	row++;

	/* Look for ENABLED static dependencies */
	col = 0;
	for (to = OMAP4_CD_WKUP; to < OMAP4_CD_NONE; to++) {
		if (statdep44xx_has_dependency_towards_it(to) != 1)
			continue;

		clkdm44xx_get_name(to, to_s);
		snprintf(table[row][0], TABLE_MAX_ELT_LEN, "  %s", to_s);
		col = 0;
		for (from = OMAP4_CD_WKUP; from < OMAP4_CD_NONE; from++) {
			if (dep44xx_register_exists(from, DEP_STATIC) != 1)
				continue;

			col++;
			if (statdep44xx_exists(from, to) != 1)
				continue;

			ret = dep44xx_is_enabled(
				from, to, DEP_STATIC);
			if (ret < 0) {
				fprintf(stderr, "%s(): "
					"dep44xx_is_enabled(%s, %s) returned "
					"with %d error!\n",
					__func__, from_s, to_s, ret);
				goto statdep44xx_show_end;
			}

			if (statdep44xx_is_read_only(from, to) == 1) {
				if (ret == 0)
					strncpy(table[row][col], "Dis (RO)",
						TABLE_MAX_ELT_LEN);
				else
					strncpy(table[row][col], "EN (RO)",
						TABLE_MAX_ELT_LEN);
			} else {
				if (ret == 0)
					strncpy(table[row][col], "Dis",
						TABLE_MAX_ELT_LEN);
				else
					strncpy(table[row][col], "EN",
						TABLE_MAX_ELT_LEN);
			}

		}
		row++;
	}

	autoadjust_table_print(table, row, max_col + 1);
	printf("NB:\n"
		"  - EN = static dependency is ENabled.\n"
		"  - Dis = static dependency is Disabled.\n"
		"  - RO = Read-Only (hard-coded, not configurable).\n"
		"  - Empty cell = static dependency does not exist.\n\n");
	ret = 0;

statdep44xx_show_end:
	return ret;
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_audit
 * @BRIEF		audit static dependencies configuration
 * @RETURNS		0 in case of success
 *			OMAPCONF_ERR_CPU
 *			OMAPCONF_ERR_ARG
 *			OMAPCONF_ERR_REG_ACCESS
 * @param[in]		stream: output file
 *			NULL: no output (silent)
 * @param[in, out]	err_nbr: pointer to return audit error number
 * @param[in, out]	wng_nbr: pointer to return audit warning number
 * @DESCRIPTION		audit static dependencies configuration
 *//*------------------------------------------------------------------------ */
int statdep44xx_audit(FILE *stream, unsigned int *err_nbr,
	unsigned int *wng_nbr)
{
	clkdm44xx_id from, to;
	int ret;
	char table[TABLE_MAX_ROW][TABLE_MAX_COL][TABLE_MAX_ELT_LEN];
	unsigned int row, col, max_col;
	char to_s[CLKDM44XX_MAX_NAME_LENGTH];
	char from_s[CLKDM44XX_MAX_NAME_LENGTH];
	int statdep_cfg, statdep_cfg_por;

	CHECK_CPU(44xx, OMAPCONF_ERR_CPU);
	CHECK_NULL_ARG(err_nbr, OMAPCONF_ERR_ARG);
	CHECK_NULL_ARG(wng_nbr, OMAPCONF_ERR_ARG);
	*err_nbr = 0;
	*wng_nbr = 0;

	/* Fill Table Header */
	autoadjust_table_init(table);
	row = 0;
	max_col = 0;
	strncpy(table[row][0], "Static Dependency AUDIT From",
		TABLE_MAX_ELT_LEN);
	for (from = OMAP4_CD_WKUP; from < OMAP4_CD_NONE; from++) {
		if (dep44xx_register_exists(from, DEP_STATIC) != 1)
			continue;

		max_col++;
		clkdm44xx_get_name(from, from_s);
		snprintf(table[row][max_col], TABLE_MAX_ELT_LEN, "%s", from_s);
	}
	row++;
	strncpy(table[row][0], "Towards", TABLE_MAX_ELT_LEN);
	row++;

	/* Browse static dependencies */
	col = 0;
	for (to = OMAP4_CD_WKUP; to < OMAP4_CD_NONE; to++) {
		if (statdep44xx_has_dependency_towards_it(to) != 1)
			continue;

		clkdm44xx_get_name(to, to_s);
		snprintf(table[row][0], TABLE_MAX_ELT_LEN, "  %s", to_s);
		col = 0;
		for (from = OMAP4_CD_WKUP; from < OMAP4_CD_NONE; from++) {
			if (dep44xx_register_exists(from, DEP_STATIC) != 1)
				continue;

			col++;
			if (statdep44xx_exists(from, to) != 1)
				continue;

			if (statdep44xx_is_read_only(from, to) == 1) {
				strncpy(table[row][col], "Ign.",
						TABLE_MAX_ELT_LEN);
				continue;
			}

			statdep_cfg = dep44xx_is_enabled(
				from, to, DEP_STATIC);
			if (statdep_cfg < 0) {
				fprintf(stderr, "%s(): "
					"dep44xx_is_enabled(%s, %s) returned "
					"with %d error!\n",
					__func__, from_s, to_s, statdep_cfg);
				goto statdep44xx_show_end;
			}
			statdep_cfg_por = statdep44xx_por_cfg_get(from, to);

			if (((statdep_cfg == 1) && (statdep_cfg_por == DEP_EN)) ||
				((statdep_cfg == 0) && (statdep_cfg_por == DEP_DIS))) {
				strncpy(table[row][col], "Pass",
					TABLE_MAX_ELT_LEN);
			} else {
				strncpy(table[row][col], "FAIL",
					TABLE_MAX_ELT_LEN);
				(*err_nbr)++;
			}
		}
		row++;
	}

	if (stream != NULL) {
		autoadjust_table_fprint(stream, table, row, max_col + 1);
		fprintf(stream, "NB:\n"
			"  - ALL STATIC DEPENDENCIES expected to be DISABLED "
				"in FINAL product (except HW errata).\n"
			"  - Pass = static dependency setting correct.\n"
			"  - FAIL = static dependency setting incorrect.\n"
			"  - Ign. = ignored as static dependency is "
				"Read-Only.\n"
			"  - Empty cell = static dependency does not exist."
				"\n\n");

		if (*err_nbr == 0) {
			fprintf(stream, "SUCCESS! Static Dependency "
				"configuration audit completed with 0 error "
				"(%d warning(s))\n\n",
				*wng_nbr);
		} else {
			fprintf(stream, "FAILED! Static Dependency "
				"configuration audit completed with %d error(s)"
				" and %d warning(s).\n\n",
				*err_nbr, *wng_nbr);
		}
	}

	ret = 0;

statdep44xx_show_end:
	return ret;
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_main_set
 * @BRIEF		set/clear given static dependency
 * @RETURNS		0 in case of success
 *			OMAPCONF_ERR_CPU
 *			OMAPCONF_ERR_ARG
 *			OMAPCONF_ERR_NOT_AVAILABLE
 *			OMAPCONF_ERR_REG_ACCESS
 * @param[in]		clkdm_from: name of the clock domain dependency is from
 *			(mandatory).
 * @param[in]		clkdm_to: name of the clock domain dependency is
 *			targeting.
 *			If clkdm_to == "all" or NULL, update all dependency from
 *			"from".
 *			List of clock domain names: "l4wkup" "emu" "mpu" "abe"
 *			"ivahd" "dsp" "alwoncore" "l4cfg" "c2c" "sdma" "mpu_m3"
 *			"l3_1" "l3_2" "l3instr" "memif" "std_efuse" "cust_efuse"
 *			"cam" "dss" "gfx" "l3init" "l4sec" "l4per" "all".
 * @param[in]		enable: set dependency if == 1, clear it if == 0.
 * @DESCRIPTION		set/clear given static dependency
 *//*------------------------------------------------------------------------ */
int statdep44xx_main_set(
	char *clkdm_from, char *clkdm_to, unsigned short enable)
{
	int ret = 0, from_found, to_found;
	clkdm44xx_id id, from, to, first, last;
	char to_s[CLKDM44XX_MAX_NAME_LENGTH];
	char from_s[CLKDM44XX_MAX_NAME_LENGTH];

	CHECK_CPU(44xx, OMAPCONF_ERR_CPU);
	CHECK_NULL_ARG(clkdm_from, OMAPCONF_ERR_ARG);
	CHECK_ARG_LESS_THAN(enable, 2, OMAPCONF_ERR_ARG);

	if (clkdm_to == NULL) {
		first = OMAP4_CD_WKUP;
		last = OMAP4_CD_NONE;
		to_found = 1;
	} else if ((strcmp(lowercase(clkdm_to), "all")) == 0) {
		first = OMAP4_CD_WKUP;
		last = OMAP4_CD_NONE;
		to_found = 1;
	} else {
		to_found = 0;
	}

	/* Get "from" & "to" clock domain IDs */
	for (from_found = 0, id = OMAP4_CD_WKUP; id < OMAP4_CD_NONE; id++) {
		if ((from_found != 1) && (strcmp(lowercase(clkdm_from),
			lowercase(clkdm44xx_get_name(id, from_s))) == 0)) {
			dprintf("%s(): found from = %s\n", __func__, from_s);
			from_found = 1;
			from = id;
		}
		if ((to_found != 1) && (strcmp(lowercase(clkdm_to),
			lowercase(clkdm44xx_get_name(id, to_s))) == 0)) {
			dprintf("%s(): found to = %s\n", __func__, to_s);
			to_found = 1;
			first = id;
			last = (clkdm44xx_id) id + 1;
		}
		if ((from_found == 1) && (to_found == 1))
			break;
	}
	if (id == OMAP4_CD_NONE) {
		dprintf("%s(): at least 1 of the clock domain IDs "
			"was not retrieved!\n",
			__func__);
		goto statdep44xx_main_set_arg_err;
	}

	if (dep44xx_register_exists(from, DEP_STATIC) != 1)
		goto statdep44xx_main_set_arg_err;

	clkdm44xx_get_name(from, from_s);

	for (to = first; to < last; to++) {
		clkdm44xx_get_name(to, to_s);
		/* Check static dependency presence, skip if doesn't exist */
		if (statdep44xx_is_read_only(from, to) != 0) {
			if (last == OMAP4_CD_NONE) {
				/* if argv[3] = "all", just skip it */
				continue;
			} else {
				if (statdep44xx_exists(from, to) != 1) {
					printf("Static Dependency from %s to %s"
						" does not exist!\n\n",
						from_s, to_s);
					goto statdep44xx_main_set_arg_err;
				} else {
					printf("Static Dependency from %s to %s"
						" is not configurable.\n\n",
						from_s, to_s);
					ret = 0;
					goto statdep44xx_main_set_end;
				}

			}
		}

		/* Update Static Dependency Setting */
		ret = statdep44xx_set(from, to, enable);
		if (ret == 0) {
			if (enable)
				printf("Static Dependency from %s to %s "
					"enabled.\n", from_s, to_s);
			else
				printf("Static Dependency from %s to %s "
					"disabled.\n", from_s, to_s);
		}
	}
	printf("\nWarning: new setting(s) applied but may be overridden "
		"unconditionally by kernel driver(s).\n\n");
	goto statdep44xx_main_set_end;

statdep44xx_main_set_arg_err:
	help(HELP_STATDEP);
	ret = OMAPCONF_ERR_ARG;

statdep44xx_main_set_end:
	return ret;
}


/* ------------------------------------------------------------------------*//**
 * @FUNCTION		statdep44xx_main
 * @BRIEF		omapconf static dependency functions entry point
 * @RETURNS		0 in case of success
 *			OMAPCONF_ERR_CPU
 *			OMAPCONF_ERR_ARG
 *			OMAPCONF_ERR_NOT_AVAILABLE
 *			OMAPCONF_ERR_REG_ACCESS
 * @param[in]		argc: number of user inputs
 * @param[in]		argv: array of user inputs
 *			Expected arguments:
 *				argc == 2 or argc == 3 or argc == 4
 *				argv[1] = "cfg", "set" or "clear"
 * @DESCRIPTION		omapconf static dependency functions entry point
 *//*------------------------------------------------------------------------ */
int statdep44xx_main(int argc, char *argv[])
{
	int ret;

	#ifdef DEP44XX_DEBUG
	int i;

	printf("%s(): argc = %d\n", __func__, argc);
	for (i = 0; i < argc; i++)
		printf("%s(): argv[%d] = %s\n", __func__, i, argv[i]);
	#endif

	if ((argc < 2) && (argc > 4))
		goto statdep44xx_main_arg_err;

	if (strcmp(argv[1], "cfg") == 0) {
		ret = statdep44xx_show();
		goto statdep44xx_main_end;
	} else if (strcmp(argv[1], "set") == 0) {
		if (argc == 3)
			ret = statdep44xx_main_set(argv[2], NULL, 1);
		else
			ret = statdep44xx_main_set(argv[2], argv[3], 1);
		goto statdep44xx_main_end;
	} else if (strcmp(argv[1], "clear") == 0) {
		if (argc == 3)
			ret = statdep44xx_main_set(argv[2], NULL, 0);
		else
			ret = statdep44xx_main_set(argv[2], argv[3], 0);
		goto statdep44xx_main_end;
	} else if (strcmp(argv[1], "dump") == 0) {
		ret = statdep44xx_dump();
		goto statdep44xx_main_end;
	}

statdep44xx_main_arg_err:
	help(HELP_STATDEP);
	ret = OMAPCONF_ERR_ARG;

statdep44xx_main_end:
	return ret;
}
